@@ -30,13 +30,12 @@
 -- @copyright Same as Nmap--See https://nmap.org/book/man-legal.html
 
 
-local bin = require "bin"
-local bit = require "bit"
 local coroutine = require "coroutine"
 local ipOps = require "ipOps"
 local nmap = require "nmap"
 local stdnse = require "stdnse"
 local string = require "string"
+local stringaux = require "stringaux"
 local table = require "table"
 local base32 = require "base32"
 local unittest = require "unittest"
@@ -307,7 +306,7 @@ end
 -- * <code>id</code>: numeric value to use for the DNS transaction id
 -- * <code>nsid</code>: If true, queries the server for the nameserver identifier (RFC 5001)
 -- * <code>subnet</code>: table, if set perform a edns-client-subnet lookup. The table should contain the fields:
---                        <code>family</code> - string can be either inet or inet6
+--                        <code>family</code> - IPv4: "inet" or 1 (default), IPv6: "inet6" or 2
 --                        <code>address</code> - string containing the originating subnet IP address
 --                        <code>mask</code> - number containing the number of subnet bits
 -- @return <code>true</code> if a dns response was received and contained an answer of the requested type,
@@ -360,9 +359,6 @@ function query(dname, options)
   if ( options.nsid ) then
     addNSID(pkt, dnssec)
   elseif ( options.subnet ) then
-    local family = { ["inet"] = 1, ["inet6"] = 2 }
-    assert( family[options.subnet.family], "Unsupported subnet family")
-    options.subnet.family = family[options.subnet.family]
     addClientSubnet(pkt, dnssec, options.subnet )
   elseif ( dnssec.DO ) then
     addOPT(pkt, {DO = true})
@@ -418,7 +414,7 @@ function reverse(ip)
     delim = ":"
     arpa = ".ip6.arpa"
   end
-  local ipParts = stdnse.strsplit(delim, ip)
+  local ipParts = stringaux.strsplit(delim, ip)
   if #ipParts == 8 then
     -- padding
     local mask = "0000"
@@ -595,7 +591,7 @@ answerFetcher[types.NSEC] = function(dec, retAll)
     return false, "No Answers"
   end
   for _, nsecrec in ipairs(nsec) do
-    table.insert( answers, ("%s:%s:%s"):format(nsecrec.name or "-", nsecrec.dname or "-", stdnse.strjoin(":", nsecrec.types) or "-"))
+    table.insert( answers, ("%s:%s:%s"):format(nsecrec.name or "-", nsecrec.dname or "-", table.concat(nsecrec.types, ":") or "-"))
   end
   if not retAll then return true, answers[1] end
   return true, answers
@@ -806,7 +802,7 @@ local function encodeFQDN(fqdn)
 
   local encQ = {}
   for part in string.gmatch(fqdn, "[^%.]+") do
-    encQ[#encQ+1] = bin.pack("p", part)
+    encQ[#encQ+1] = string.pack("s1", part)
   end
   encQ[#encQ+1] = "\0"
   return table.concat(encQ)
@@ -821,7 +817,7 @@ local function encodeQuestions(questions)
   local encQ = {}
   for _, v in ipairs(questions) do
     encQ[#encQ+1] = encodeFQDN(v.dname)
-    encQ[#encQ+1] = bin.pack(">SS", v.dtype, v.class)
+    encQ[#encQ+1] = string.pack(">I2I2", v.dtype, v.class)
   end
   return table.concat(encQ)
 end
@@ -839,7 +835,7 @@ local function encodeUpdates(updates)
   local encQ = {}
   for _, v in ipairs(updates) do
     encQ[#encQ+1] = encodeFQDN(v.dname)
-    encQ[#encQ+1] = bin.pack(">SSISA", v.dtype, v.class, v.ttl, #v.data, v.data)
+    encQ[#encQ+1] = string.pack(">I2I2I4s2", v.dtype, v.class, v.ttl, v.data)
   end
   return table.concat(encQ)
 end
@@ -847,14 +843,14 @@ end
 ---
 -- Encodes the additional part of a DNS request.
 -- @param additional Table of additional records. Each must have the keys
--- <code>type</code>, <code>class</code>, <code>ttl</code>, <code>rdlen</code>,
+-- <code>type</code>, <code>class</code>, <code>ttl</code>,
 -- and <code>rdata</code>.
 -- @return Encoded additional string.
 local function encodeAdditional(additional)
   if type(additional) ~= "table" then return nil end
   local encA = {}
   for _, v in ipairs(additional) do
-    encA[#encA+1] = bin.pack(">xSSISA",  v.type, v.class, v.ttl, v.rdlen, v.rdata)
+    encA[#encA+1] = string.pack(">xI2I2I4s2",  v.type, v.class, v.ttl, v.rdata)
   end
   return table.concat(encA)
 end
@@ -865,23 +861,22 @@ end
 -- RA, RCx).
 -- @return Binary digit string representing flags.
 local function encodeFlags(flags)
-  if type(flags) == "string" then return flags end
+  if type(flags) == "number" then return flags end
   if type(flags) ~= "table" then return nil end
-  local fb = ""
-  if flags.QR then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.OC1 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.OC2 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.OC3 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.OC4 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.AA then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.TC then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.RD then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.RA then fb = fb .. "1" else fb = fb .. "0" end
-  fb = fb .. "000"
-  if flags.RC1 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.RC2 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.RC3 then fb = fb .. "1" else fb = fb .. "0" end
-  if flags.RC4 then fb = fb .. "1" else fb = fb .. "0" end
+  local fb = 0
+  if flags.QR  then fb = fb|0x8000 end
+  if flags.OC1 then fb = fb|0x4000 end
+  if flags.OC2 then fb = fb|0x2000 end
+  if flags.OC3 then fb = fb|0x1000 end
+  if flags.OC4 then fb = fb|0x0800 end
+  if flags.AA  then fb = fb|0x0400 end
+  if flags.TC  then fb = fb|0x0200 end
+  if flags.RD  then fb = fb|0x0100 end
+  if flags.RA  then fb = fb|0x0080 end
+  if flags.RC1 then fb = fb|0x0008 end
+  if flags.RC2 then fb = fb|0x0004 end
+  if flags.RC3 then fb = fb|0x0002 end
+  if flags.RC4 then fb = fb|0x0001 end
   return fb
 end
 
@@ -912,9 +907,9 @@ function encode(pkt)
 
   local encStr
   if ( pkt.flags.raw ) then
-    encStr = bin.pack(">SSS4", pkt.id, pkt.flags.raw, qorzlen, aorplen, aorulen, #pkt.additional) .. data .. additional
+    encStr = string.pack(">I2I2I2I2I2I2", pkt.id, pkt.flags.raw, qorzlen, aorplen, aorulen, #pkt.additional) .. data .. additional
   else
-    encStr = bin.pack(">SBS4", pkt.id, encFlags, qorzlen, aorplen, aorulen, #pkt.additional) .. data .. additional
+    encStr = string.pack(">I2I2I2I2I2I2", pkt.id, encFlags, qorzlen, aorplen, aorulen, #pkt.additional) .. data .. additional
   end
   return encStr
 end
@@ -928,6 +923,7 @@ end
 -- @return Decoded domain, or <code>nil</code> on error.
 function decStr(data, pos)
   local function dec(data, pos, limit)
+    assert(pos > 0)
     local partlen
     local parts = {}
     local part
@@ -938,17 +934,17 @@ function decStr(data, pos)
       return pos, nil
     end
 
-    pos, partlen = bin.unpack(">C", data, pos)
+    partlen, pos = string.unpack(">B", data, pos)
     while (partlen ~= 0) do
       if (partlen < 64) then
-        pos, part = bin.unpack("A" .. partlen, data, pos)
-        if part == nil then
+        if (#data - pos + 1) < partlen then
           return pos
         end
+        part, pos = string.unpack("c" .. partlen, data, pos)
         table.insert(parts, part)
-        pos, partlen = bin.unpack(">C", data, pos)
+        partlen, pos = string.unpack(">B", data, pos)
       else
-        pos, partlen = bin.unpack(">S", data, pos - 1)
+        partlen, pos = string.unpack(">I2", data, pos - 1)
         local _, part = dec(data, partlen - 0xC000 + 1, limit - 1)
         if part == nil then
           return pos
@@ -976,7 +972,7 @@ local function decodeQuestions(data, count, pos)
   for i = 1, count do
     local currQ = {}
     pos, currQ.dname = decStr(data, pos)
-    pos, currQ.dtype, currQ.class = bin.unpack(">SS", data, pos)
+    currQ.dtype, currQ.class, pos = string.unpack(">I2I2", data, pos)
     table.insert(q, currQ)
   end
   return pos, q
@@ -990,23 +986,13 @@ local decoder = {}
 -- Decodes IP of A record, puts it in <code>entry.ip</code>.
 -- @param entry RR in packet.
 decoder[types.A] = function(entry)
-  local ip = {}
-  local _
-  _, ip[1], ip[2], ip[3], ip[4] = bin.unpack(">C4", entry.data)
-  entry.ip = table.concat(ip, ".")
+  entry.ip = ipOps.str_to_ip(entry.data:sub(1,4))
 end
 
 -- Decodes IP of AAAA record, puts it in <code>entry.ipv6</code>.
 -- @param entry RR in packet.
 decoder[types.AAAA] = function(entry)
-  local ip = {}
-  local pos = 1
-  local num
-  for i = 1, 8 do
-    pos, num = bin.unpack(">S", entry.data, pos)
-    table.insert(ip, string.format('%x', num))
-  end
-  entry.ipv6 = table.concat(ip, ":")
+  entry.ipv6 = ipOps.str_to_ip(entry.data:sub(1,16))
 end
 
 -- Decodes SSH fingerprint record, puts it in <code>entry.SSHFP</code> as
@@ -1016,10 +1002,10 @@ end
 -- <code>fptype</code>, and <code>fingerprint</code>.
 -- @param entry RR in packet.
 decoder[types.SSHFP] = function(entry)
-  local _
+  local pos
   entry.SSHFP = {}
-  _, entry.SSHFP.algorithm,
-  entry.SSHFP.fptype, entry.SSHFP.fingerprint = bin.unpack(">C2H" .. (#entry.data - 2), entry.data)
+  entry.SSHFP.algorithm, entry.SSHFP.fptype, pos = string.unpack(">BB", entry.data)
+  entry.SSHFP.fingerprint = stdnse.tohex(entry.data:sub(pos))
 end
 
 
@@ -1039,12 +1025,12 @@ decoder[types.SOA] = function(entry, data, pos)
 
   np, entry.SOA.mname = decStr(data, np)
   np, entry.SOA.rname = decStr(data, np)
-  np, entry.SOA.serial,
+  entry.SOA.serial,
   entry.SOA.refresh,
   entry.SOA.retry,
   entry.SOA.expire,
-  entry.SOA.minimum
-  = bin.unpack(">I5", data, np)
+  entry.SOA.minimum,
+  np = string.unpack(">I4I4I4I4I4", data, np)
 end
 
 -- An iterator that returns the positions of nonzero bits in the given binary
@@ -1057,11 +1043,11 @@ local function bit_iter(bits)
       local mask = 0x80
 
       while mask > 0 do
-        if bit.band(n, mask) ~= 0 then
+        if (n & mask) ~= 0 then
           coroutine.yield((i - 1) * 8 + j)
         end
         j = j + 1
-        mask = bit.rshift(mask, 1)
+        mask = (mask >> 1)
       end
     end
   end)
@@ -1082,7 +1068,7 @@ decoder[types.NSEC] = function (entry, data, pos)
   np, entry.NSEC.next_dname = decStr(data, np)
   while np < pos do
     local block_num, type_bitmap
-    np, block_num, type_bitmap = bin.unpack(">Cp", data, np)
+    block_num, type_bitmap, np = string.unpack(">Bs1", data, np)
     entry.NSEC.types = {}
     for i in bit_iter(type_bitmap) do
       entry.NSEC.types[(block_num - 1) * 256 + i] = true
@@ -1108,28 +1094,21 @@ decoder[types.NSEC3] = function (entry, data, pos)
   entry.NSEC3.dname = entry.dname
   entry.NSEC3.salt, entry.NSEC3.hash = {}, {}
 
-  np, entry.NSEC3.hash.alg,flags,entry.NSEC3.iterations = bin.unpack(">CBS", data, np)
+  entry.NSEC3.hash.alg, entry.NSEC3.flags, entry.NSEC3.iterations, np = string.unpack(">BBI2", data, np)
   -- do we even need to decode these do we care about opt out?
   -- entry.NSEC3.flags = decodeFlagsNSEC3(flags)
 
-  np, entry.NSEC3.salt.bin = bin.unpack(">p",  data, np)
-  _, entry.NSEC3.salt.hex = bin.unpack("H" .. #entry.NSEC3.salt.bin, entry.NSEC3.salt.bin)
+  entry.NSEC3.salt.bin, np = string.unpack(">s1",  data, np)
+  entry.NSEC3.salt.hex = stdnse.tohex(entry.NSEC3.salt.bin)
 
-  np, entry.NSEC3.hash.bin = bin.unpack(">p" , data, np)
-  _, entry.NSEC3.hash.hex = bin.unpack(">H" .. #entry.NSEC3.hash.bin , entry.NSEC3.hash.bin)
+  entry.NSEC3.hash.bin, np = string.unpack(">s1" , data, np)
+  entry.NSEC3.hash.hex = stdnse.tohex(entry.NSEC3.hash.bin)
   entry.NSEC3.hash.base32 = base32.enc(entry.NSEC3.hash.bin, true)
 
-  np, entry.NSEC3.WinBlockNo, entry.NSEC3.bmplength = bin.unpack(">CC", data, np)
-  np, entry.NSEC3.bin = bin.unpack(">B".. entry.NSEC3.bmplength, data, np)
+  entry.NSEC3.WinBlockNo, entry.NSEC3.bin, np = string.unpack(">Bs1", data, np)
   entry.NSEC3.types = {}
-  if entry.NSEC3.bin == nil then
-    entry.NSEC3.bin = ""
-  end
-  for i=1, string.len(entry.NSEC3.bin) do
-    local bit = string.sub(entry.NSEC3.bin,i,i)
-    if bit == "1" then
-      table.insert(entry.NSEC3.types, (entry.NSEC3.WinBlockNo*256+i-1))
-    end
+  for i in bit_iter(entry.NSEC3.bin) do
+    entry.NSEC3.types[(entry.NSEC3.WinBlockNo - 1) * 256 + i] = true
   end
 end
 
@@ -1181,9 +1160,8 @@ decoder[types.PTR] = decDomain
 decoder[types.TXT] =
 function (entry, data, pos)
 
-  local len = entry.data:len()
-  local np = pos - #entry.data
-  local txt_len
+  local len = #entry.data
+  local np = pos - len
   local txt
 
   if len > 0 then
@@ -1191,10 +1169,8 @@ function (entry, data, pos)
     entry.TXT.text = {}
   end
 
-  while len > 0 do
-    np, txt_len = bin.unpack("C", data, np)
-    np, txt = bin.unpack("A" .. txt_len, data, np )
-    len = len - txt_len - 1
+  while np < pos do
+    txt, np = string.unpack("s1", data, np)
     table.insert( entry.TXT.text, txt )
   end
 
@@ -1213,8 +1189,7 @@ decoder[types.OPT] =
 function(entry, data, pos)
   local np = pos - #entry.data - 6
   local opt = { bufsize = entry.class }
-  np, opt.rcode, opt.version, opt.zflags, opt.rdlen = bin.unpack(">CCSS", data, np)
-  np, opt.data = bin.unpack("A" .. opt.rdlen, data, np)
+  opt.rcode, opt.version, opt.zflags, opt.data, np = string.unpack(">BBI2s2", data, np)
   entry.OPT = opt
 end
 
@@ -1231,7 +1206,7 @@ function(entry, data, pos)
   local np = pos - #entry.data + 2
   local _
   entry.MX = {}
-  _, entry.MX.pref = bin.unpack(">S", entry.data)
+  entry.MX.pref = string.unpack(">I2", entry.data)
   _, entry.MX.server = decStr(data, np)
 end
 
@@ -1248,7 +1223,7 @@ function(entry, data, pos)
   local np = pos - #entry.data
   local _
   entry.SRV = {}
-  np, entry.SRV.prio, entry.SRV.weight, entry.SRV.port = bin.unpack(">S>S>S", data, np)
+  entry.SRV.prio, entry.SRV.weight, entry.SRV.port, np = string.unpack(">I2I2I2", data, np)
   np, entry.SRV.target = decStr(data, np)
 end
 
@@ -1262,12 +1237,9 @@ local function decodeRR(data, count, pos)
   for i = 1, count do
     local currRR = {}
     pos, currRR.dname = decStr(data, pos)
-    pos, currRR.dtype, currRR.class, currRR.ttl = bin.unpack(">SSI", data, pos)
+    currRR.dtype, currRR.class, currRR.ttl, pos = string.unpack(">I2I2I4", data, pos)
 
-    local reslen
-    pos, reslen = bin.unpack(">S", data, pos)
-
-    pos, currRR.data = bin.unpack("A" .. reslen, data, pos)
+    currRR.data, pos = string.unpack(">s2", data, pos)
 
     -- try to be smart: decode per type
     if decoder[currRR.dtype] then
@@ -1279,39 +1251,26 @@ local function decodeRR(data, count, pos)
   return pos, ans
 end
 
----
--- Splits a string up into a table of single characters.
--- @param str String to be split up.
--- @return Table of characters.
-local function str2tbl(str)
-  local tbl = {}
-  for i = 1, #str do
-    table.insert(tbl, string.sub(str, i, i))
-  end
-  return tbl
-end
-
 ---
 -- Decodes DNS flags.
 -- @param flgStr Flags as a binary digit string.
 -- @return Table representing flags.
-local function decodeFlags(flgStr)
-  local flags = {}
-  local flgTbl = str2tbl(flgStr)
-  if flgTbl[1] == '1' then flags.QR = true end
-  if flgTbl[2] == '1' then flags.OC1 = true end
-  if flgTbl[3] == '1' then flags.OC2 = true end
-  if flgTbl[4] == '1' then flags.OC3 = true end
-  if flgTbl[5] == '1' then flags.OC4 = true end
-  if flgTbl[6] == '1' then flags.AA = true end
-  if flgTbl[7] == '1' then flags.TC = true end
-  if flgTbl[8] == '1' then flags.RD = true end
-  if flgTbl[9] == '1' then flags.RA = true end
-  if flgTbl[13] == '1' then flags.RC1 = true end
-  if flgTbl[14] == '1' then flags.RC2 = true end
-  if flgTbl[15] == '1' then flags.RC3 = true end
-  if flgTbl[16] == '1' then flags.RC4 = true end
-  return flags
+local function decodeFlags(flags)
+  local tflags = {}
+  if (flags & 0x8000) ~= 0 then tflags.QR  = true end
+  if (flags & 0x4000) ~= 0 then tflags.OC1 = true end
+  if (flags & 0x2000) ~= 0 then tflags.OC2 = true end
+  if (flags & 0x1000) ~= 0 then tflags.OC3 = true end
+  if (flags & 0x0800) ~= 0 then tflags.OC4 = true end
+  if (flags & 0x0400) ~= 0 then tflags.AA  = true end
+  if (flags & 0x0200) ~= 0 then tflags.TC  = true end
+  if (flags & 0x0100) ~= 0 then tflags.RD  = true end
+  if (flags & 0x0080) ~= 0 then tflags.RA  = true end
+  if (flags & 0x0008) ~= 0 then tflags.RC1 = true end
+  if (flags & 0x0004) ~= 0 then tflags.RC2 = true end
+  if (flags & 0x0002) ~= 0 then tflags.RC3 = true end
+  if (flags & 0x0001) ~= 0 then tflags.RC4 = true end
+  return tflags
 end
 
 ---
@@ -1323,7 +1282,7 @@ function decode(data)
   local pkt = {}
   local encFlags
   local cnt = {}
-  pos, pkt.id, encFlags, cnt.q, cnt.a, cnt.auth, cnt.add = bin.unpack(">SB2S4", data)
+  pkt.id, encFlags, cnt.q, cnt.a, cnt.auth, cnt.add, pos = string.unpack(">I2I2I2I2I2I2", data)
   -- for now, don't decode the flags
   pkt.flags = decodeFlags(encFlags)
 
@@ -1332,8 +1291,9 @@ function decode(data)
   -- a quick fix to allow decoding of non updates and not break for updates
   -- the flags are enough for the current code to determine whether an update was successful or not
   --
-  local strflags=encodeFlags(pkt.flags)
-  if ( strflags:sub(1,4) == "1010" ) then
+  local flags = encodeFlags(pkt.flags)
+  -- QR, OC2
+  if (flags & 0xF000) == 0xA000 then
     return pkt
   else
     pos, pkt.questions = decodeQuestions(data, cnt.q, pos)
@@ -1401,33 +1361,39 @@ end
 -- @param flags Flag table, each entry representing a flag (only DO flag implmented).
 -- @return Binary digit string representing flags.
 local function encodeOPT_Z(flags)
-  if type(flags) == "string" then return flags end
-  if type(flags) ~= "table" then return nil end
-  local bits = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
-  for n, k in pairs({[1] = "DO"}) do
-    if flags[k] then
-      bits[n] = 1
-    end
-  end
-  return table.concat(bits)
+  if type(flags) == "number" then return flags end
+  assert(type(flags) == "table")
+  local bits = 0
+  if flags.DO then bits = bits|0x8000 end
+  return bits
 end
 
 ---
 -- Adds an client-subnet payload to the OPT packet
 --
--- implementing http://tools.ietf.org/html/draft-vandergaast-edns-client-subnet-00
+-- implementing https://tools.ietf.org/html/rfc7871
 -- @param pkt Table representing DNS packet.
 -- @param Z Table of Z flags. Only DO is supported.
 -- @param client_subnet table containing the following fields
---        <code>family</code>  - 1 IPv4, 2 - IPv6
+--        <code>family</code>  - IPv4: "inet" or 1 (default), IPv6: "inet6" or 2
 --        <code>mask</code>    - byte containing the length of the subnet mask
 --        <code>address</code> - string containing the IP address
 function addClientSubnet(pkt,Z,subnet)
-  local udp_payload_size = 4096
-  local code = 20730 -- temporary option-code http://comments.gmane.org/gmane.ietf.dnsext/19776
-  local scope_mask = 0 -- In requests, it MUST be set to 0 see draft
-  local data = bin.pack(">SCCA",subnet.family or 1,subnet.mask,scope_mask,ipOps.ip_to_str(subnet.address))
-  local opt = bin.pack(">SS",code, #data) .. data
+  local family = subnet.family or 1
+  if type(family) == "string" then
+    family = ({inet=1,inet6=2})[family]
+  end
+  assert(family == 1 or family == 2, "Unsupported subnet family")
+  local code = 8 -- https://www.iana.org/assignments/dns-parameters/dns-parameters.xhtml#dns-parameters-11
+  local mask = subnet.mask
+  local scope_mask = 0 -- In requests, it MUST be set to 0
+  -- Per RFC 7871, section 6:
+  -- Address must have all insignificant bits zeroed out and insignificant bytes
+  -- must be trimmed off. (/24 IPv4 address is submitted as 3 octets, not 4.)
+  local addr = ipOps.get_first_ip(subnet.address, mask)
+  addr = ipOps.ip_to_str(addr):sub(1, (mask + 7) // 8)
+  local data = string.pack(">I2BB", family, mask, scope_mask) .. addr
+  local opt = string.pack(">I2s2", code, data)
   addOPT(pkt,Z,opt)
 end
 
@@ -1436,8 +1402,7 @@ end
 -- @param pkt Table representing DNS packet.
 -- @param Z Table of Z flags. Only DO is supported.
 function addNSID (pkt,Z)
-  local udp_payload_size = 4096
-  local opt = bin.pack(">SS",3, 0) -- nsid data
+  local opt = string.pack(">I2I2", 3, 0) -- nsid data
   addOPT(pkt,Z,opt)
 end
 
@@ -1452,12 +1417,11 @@ function addOPT(pkt, Z, opt)
   local rdata = opt or ""
   if type(pkt) ~= "table" then return nil end
   if type(pkt.additional) ~= "table" then return nil end
-  local _, Z_int = bin.unpack(">S", bin.pack("B", encodeOPT_Z(Z)))
+  local Z_int = encodeOPT_Z(Z)
   local opt = {
     type = types.OPT,
     class = 4096,  -- Actually the sender UDP payload size.
     ttl = 0 * (0x01000000) + 0 * (0x00010000) + Z_int,
-    rdlen = #rdata,
     rdata = rdata,
   }
   table.insert(pkt.additional, opt)
@@ -1534,8 +1498,8 @@ function update(dname, options)
   flags.RD = false
   flags.OC1, flags.OC2, flags.OC3, flags.OC4 = false, true, false, true
 
-  -- If ttl is zero and updata is string and zero length or nil, assume delete record
-  if ( ttl == 0 and ( ( type(updata) == "string" and #updata == 0 ) or not(updata) ) ) then
+  -- If ttl is zero and updata is nil or a string of zero length, assume delete record
+  if ttl == 0 and (not updata or (type(updata) == "string" and #updata == 0)) then
     class = CLASS.ANY
     updata = ""
     if ( types.MX == dtype and not(options.zone) ) then zone=dname end
@@ -1544,19 +1508,19 @@ function update(dname, options)
     end
     -- if not, let's try to update the zone
   else
-    if ( dtype == types.A ) then
-      updata = updata and bin.pack(">I", ipOps.todword(updata)) or ""
+    if ( dtype == types.A or dtype == types.AAAA ) then
+      updata = updata and ipOps.ip_to_str(updata) or ""
     elseif( dtype == types.CNAME ) then
       updata = encodeFQDN(updata)
     elseif( dtype == types.MX ) then
       assert( not( type(updata) ~= "table" ), "dns.update expected options.data to be a table")
       if ( not(options.zone) ) then zone = dname end
-      local data = bin.pack(">S", updata.pref)
+      local data = string.pack(">I2", updata.pref)
       data = data .. encodeFQDN(updata.mx)
       updata = data
     elseif ( dtype == types.SRV ) then
       assert( not( type(updata) ~= "table" ), "dns.update expected options.data to be a table")
-      local data = bin.pack(">SSS", updata.prio, updata.weight, updata.port )
+      local data = string.pack(">I2I2I2", updata.prio, updata.weight, updata.port )
       data = data .. encodeFQDN(updata.target)
       updata = data
       zone = options.zone or dname:match("^_.-%._.-%.(.+)$")
@@ -1573,8 +1537,8 @@ function update(dname, options)
 
   if ( status ) then
     local decoded = decode(response[1].data)
-    local flags=encodeFlags(decoded.flags)
-    if (flags:sub(-4) == "0000") then
+    local flags = encodeFlags(decoded.flags)
+    if (flags & 0xF) == 0 then
       return true
     end
   end
@@ -1589,4 +1553,285 @@ end
 test_suite = unittest.TestSuite:new()
 
 test_suite:add_test(unittest.equal(encodeFQDN("test.me.com"), "\x04test\x02me\x03com\0"), "encodeFQDN")
+local tests = { {
+    pkt = string.char(0x92, 0xdc, -- Trsnsaction ID
+      0x81, 0x80, -- Flags
+      0x00, 0x01, -- Questions count
+      0x00, 0x01, -- Answers RRs count
+      0x00, 0x00, -- Authorities RRs count
+      0x00, 0x00, -- Additionals RRs count
+      0x06, -- Label length <-- [12]
+      0x73, 0x63, 0x61, 0x6e, 0x6d, 0x65, -- "scanme"
+      0x04, -- Label length
+      0x6e, 0x6d, 0x61, 0x70, -- "nmap"
+      0x03, -- Label length
+      0x6f, 0x72, 0x67, -- "org"
+      0x00, -- Name terminator
+      0x00, 0x01, -- A
+      0x00, 0x01, -- CLASS_IN
+      0xc0, 0x0c, -- Compressed name pointer to offset 12
+      0x00, 0x01, -- A
+      0x00, 0x01, -- CLASS_IN
+      0x00, 0x00, 0x0e, 0x0f, -- TTL 3599
+      0x00, 0x04, -- Record Length
+      0x2d, 0x21, 0x20, 0x9c ), -- 45.33.32.156
+    qname = "scanme.nmap.org",
+    qtype = "A",
+    ans = {"ip", "45.33.32.156"}
+  },
+  {
+    pkt = string.char(
+      0x08, 0xf2, -- ID
+      0x81, 0x80, -- Flags
+      0x00, 0x01, -- Questions count
+      0x00, 0x01, -- Answers RRs count
+      0x00, 0x00, -- Authorities RRs count
+      0x00, 0x00, -- Additionals RRs count
+      0x03, -- Label length
+      0x31, 0x35, 0x36, -- "156"
+      0x02, -- Label length
+      0x33, 0x32, -- "32"
+      0x02, -- Label length
+      0x33, 0x33, -- "33"
+      0x02, -- Label length
+      0x34, 0x35, -- "45"
+      0x07, -- Label length
+      0x69, 0x6e, 0x2d, 0x61, 0x64, 0x64, 0x72, -- "in-addr"
+      0x04, -- Label length
+      0x61, 0x72, 0x70, 0x61, -- "arpa"
+      0x00, -- Name terminator
+      0x00, 0x0c, -- PTR
+      0x00, 0x01, -- CLASS_IN
+      0xc0, 0x0c, -- Compressed name pointer to offset 12
+      0x00, 0x0c, -- PTR
+      0x00, 0x01, -- CLASS_IN
+      0x00, 0x01, 0x51, 0x78, -- TTL 86392
+      0x00, 0x11, -- Record Length
+      0x06, -- Label length
+      0x73, 0x63, 0x61, 0x6e, 0x6d, 0x65, -- "scanme"
+      0x04, -- Label length
+      0x6e, 0x6d, 0x61, 0x70, -- "nmap"
+      0x03, -- Label length
+      0x6f, 0x72, 0x67, -- "org"
+      0x00),  -- Name terminator
+    qname = "156.32.33.45.in-addr.arpa",
+    qtype = "PTR",
+    ans = {"domain", "scanme.nmap.org"}
+  },
+}
+for _, t in ipairs(tests) do
+  local decoded = decode(t.pkt)
+  local q = decoded.questions[1]
+  local a = decoded.answers[1]
+  test_suite:add_test(unittest.equal(q.dname, t.qname), "decode question name")
+  test_suite:add_test(unittest.equal(q.dtype, types[t.qtype]), "decode question type")
+  test_suite:add_test(unittest.equal(a[t.ans[1]], t.ans[2]), "decode answer")
+end
+local axfr = stdnse.fromhex(
+"dead840000010032000000000c7a\z
+6f6e657472616e73666572026d650000\z
+fc0001c00c0006000100001c20002f06\z
+6e737a746d310464696769056e696e6a\z
+610005726f62696ec034785908810002\z
+a300000003840012750000000e10c00c\z
+000d00010000012c00190d436173696f\z
+2066782d373030470a57696e646f7773\z
+205850c00c001000010000012d004544\z
+676f6f676c652d736974652d76657269\z
+6669636174696f6e3d74795032384a37\z
+4a41554841396677327348584d676343\z
+4330493658426d6d6f56693034566c4d\z
+65777841c00c000f000100001c200016\z
+0000054153504d58014c06474f4f474c\z
+4503434f4d00c00c000f000100001c20\z
+0009000a04414c5431c0e0c00c000f00\z
+0100001c200009000a04414c5432c0e0\z
+c00c000f000100001c20001600140641\z
+53504d58320a474f4f474c454d41494c\z
+c0efc00c000f000100001c20000b0014\z
+064153504d5833c133c00c000f000100\z
+001c20000b0014064153504d5834c133\z
+c00c000f000100001c20000b00140641\z
+53504d5835c133c00c0001000100001c\z
+20000405c4690ec00c0002000100001c\z
+200002c02dc00c0002000100001c2000\z
+09066e737a746d32c0340f5f61636d65\z
+2d6368616c6c656e6765c00c00100001\z
+0000012d002c2b364f6130356862554a\z
+39785373765979377041705176774355\z
+535347677876726264697a6a65504573\z
+5a49045f736970045f746370c00c0021\z
+0001000036b0001b0000000013c40377\z
+77770c7a6f6e657472616e7366657202\z
+6d650002313403313035033139360135\z
+07494e2d414444520441525041c22000\z
+0c000100001c200002c21c0c61736664\z
+6261757468646e73c220001200010000\z
+1edc001c0001086173666462626f780c\z
+7a6f6e657472616e73666572026d6500\z
+c2740001000100001c2000047f000001\z
+0b6173666462766f6c756d65c27d0012\z
+000100001e78001c0001086173666462\z
+626f780c7a6f6e657472616e73666572\z
+026d65000f63616e62657272612d6f66\z
+66696365c2c10001000100001c200004\z
+ca0e51e607636d6465786563c2c10010\z
+00010000012c0005043b206c7307636f\z
+6e74616374c2c10010000100278d0000\z
+646352656d656d62657220746f206361\z
+6c6c206f7220656d61696c2050697070\z
+61206f6e202b34342031323320343536\z
+37383930206f72207069707061407a6f\z
+6e657472616e736665722e6d65207768\z
+656e206d616b696e6720444e53206368\z
+616e6765730964632d6f6666696365c2\z
+c10001000100001c2000048fe4b58408\z
+6465616462656566c2c1001c00010000\z
+1c210010deadbeaf0000000000000000\z
+00000000026472c2c1001d0001000001\z
+2c0010001216138b728cee7fa5c44a00\z
+98968003445a43c2c10010000100001c\z
+200008074162436445664705656d6169\z
+6cc2c100230001000008ae0038000100\z
+010150094532552b656d61696c000565\z
+6d61696c0c7a6f6e657472616e736665\z
+72026d650c7a6f6e657472616e736665\z
+72026d6500c3f90001000100001c2000\z
+044a7dce1a0548656c6c6fc432001000\z
+0100001c20001d1c486920746f204a6f\z
+736820616e6420616c6c206869732063\z
+6c61737304686f6d65c4320001000100\z
+001c2000047f00000104496e666fc432\z
+0010000100001c20008b8a5a6f6e6554\z
+72616e736665722e6d65207365727669\z
+63652070726f76696465642062792052\z
+6f62696e20576f6f64202d20726f6269\z
+6e40646967692e6e696e6a612e205365\z
+6520687474703a2f2f646967692e6e69\z
+6e6a612f70726f6a656374732f7a6f6e\z
+657472616e736665726d652e70687020\z
+666f72206d6f726520696e666f726d61\z
+74696f6e2e08696e7465726e616cc432\z
+000200010000012c000906696e746e73\z
+31c432c533000200010000012c000906\z
+696e746e7332c432c548000100010000\z
+012c000451046c29c55d000100010000\z
+012c0004a7582a5e066f6666696365c4\z
+320001000100001c200004041727fe0a\z
+697076366163746e6f77036f7267c432\z
+001c000100001c2000102001067c02e8\z
+001100000000c1001332036f7761c432\z
+0001000100001c200004cf2ec5200972\z
+6f62696e776f6f64c432001000010000\z
+012e000b0a526f62696e20576f6f6402\z
+7270c432001100010000014100320572\z
+6f62696e0c7a6f6e657472616e736665\z
+72026d650009726f62696e776f6f640c\z
+7a6f6e657472616e73666572026d6500\z
+03736970c62d0023000100000d05003b\z
+000200030150074532552b7369702b21\z
+5e2e2a24217369703a637573746f6d65\z
+722d73657276696365407a6f6e657472\z
+616e736665722e6d6521000473716c69\z
+c62d001000010000012c000c0b27206f\z
+7220313d31202d2d067373686f636bc6\z
+2d0010000100001c20001c1b2829207b\z
+203a5d7d3b206563686f205368656c6c\z
+53686f636b65640773746167696e67c6\z
+2d0005000100001c20001a0377777710\z
+7379646e65796f70657261686f757365\z
+03636f6d000f616c6c746370706f7274\z
+736f70656e086669726577616c6c0474\z
+657374c62d000100010000012d00047f\z
+0000010774657374696e67c62d000500\z
+010000012d0002c21c0376706ec62d00\z
+01000100000fa00004ae243b9ac21c00\z
+01000100001c20000405c4690e037873\z
+73c62d001000010000012c00201f273e\z
+3c7363726970743e616c657274282742\z
+6f6f27293c2f7363726970743ec62d00\z
+06000100001c200018c02dc040785908\z
+810002a300000003840012750000000e\z
+10")
+local axfr_decoded = decode(axfr)
+local answers = {
+  {dname="zonetransfer.me", dtype=types["SOA"], mname="nsztm1.digi.ninja", rname="robin.digi.ninja"},
+  {dname="zonetransfer.me", dtype=types["HINFO"]}, --  "Casio fx-700G" "Windows XP"
+  {dname="zonetransfer.me", dtype=types["TXT"], text="google-site-verification=tyP28J7JAUHA9fw2sHXMgcCC0I6XBmmoVi04VlMewxA"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=0,  server="ASPMX.L.GOOGLE.COM"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=10, server="ALT1.ASPMX.L.GOOGLE.COM"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=10, server="ALT2.ASPMX.L.GOOGLE.COM"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=20, server="ASPMX2.GOOGLEMAIL.COM"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=20, server="ASPMX3.GOOGLEMAIL.COM"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=20, server="ASPMX4.GOOGLEMAIL.COM"},
+  {dname="zonetransfer.me", dtype=types["MX"], pref=20, server="ASPMX5.GOOGLEMAIL.COM"},
+  {dname="zonetransfer.me", dtype=types["A"],  ip="5.196.105.14"},
+  {dname="zonetransfer.me", dtype=types["NS"], domain="nsztm1.digi.ninja"},
+  {dname="zonetransfer.me", dtype=types["NS"], domain="nsztm2.digi.ninja"},
+  {dname="_acme-challenge.zonetransfer.me", dtype=types["TXT"], text="6Oa05hbUJ9xSsvYy7pApQvwCUSSGgxvrbdizjePEsZI"},
+  {dname="_sip._tcp.zonetransfer.me", dtype=types["SRV"], prio=0, weight=0, port=5060, target="www.zonetransfer.me"},
+  {dname="14.105.196.5.IN-ADDR.ARPA.zonetransfer.me", dtype=types["PTR"], domain="www.zonetransfer.me"},
+  {dname="asfdbauthdns.zonetransfer.me", dtype=types["AFSDB"]}, -- 1 asfdbbox.zonetransfer.me.
+  {dname="asfdbbox.zonetransfer.me", dtype=types["A"], ip="127.0.0.1"},
+  {dname="asfdbvolume.zonetransfer.me", dtype=types["AFSDB"]}, -- 1 asfdbbox.zonetransfer.me.
+  {dname="canberra-office.zonetransfer.me", dtype=types["A"], ip="202.14.81.230"},
+  {dname="cmdexec.zonetransfer.me", dtype=types["TXT"], text="; ls"},
+  {dname="contact.zonetransfer.me", dtype=types["TXT"], text="Remember to call or email Pippa on +44 123 4567890 or pippa@zonetransfer.me when making DNS changes"},
+  {dname="dc-office.zonetransfer.me", dtype=types["A"], ip="143.228.181.132"},
+  {dname="deadbeef.zonetransfer.me", dtype=types["AAAA"], ipv6="dead:beaf::"},
+  {dname="dr.zonetransfer.me", dtype=types["LOC"]}, -- 53.349044 N 1.642646 W 0m 1.0m 10000.0m 10.0m
+  {dname="DZC.zonetransfer.me", dtype=types["TXT"], text="AbCdEfG"},
+  {dname="email.zonetransfer.me", dtype=types["NAPTR"]}, -- 1 1 "P" "E2U+email" "" email.zonetransfer.me.zonetransfer.me.
+  {dname="email.zonetransfer.me", dtype=types["A"], ip="74.125.206.26"},
+  {dname="Hello.zonetransfer.me", dtype=types["TXT"], text="Hi to Josh and all his class"},
+  {dname="home.zonetransfer.me", dtype=types["A"], ip="127.0.0.1"},
+  {dname="Info.zonetransfer.me", dtype=types["TXT"], text="ZoneTransfer.me service provided by Robin Wood - robin@digi.ninja. See http://digi.ninja/projects/zonetransferme.php for more information."},
+  {dname="internal.zonetransfer.me", dtype=types["NS"], domain="intns1.zonetransfer.me"},
+  {dname="internal.zonetransfer.me", dtype=types["NS"], domain="intns2.zonetransfer.me"},
+  {dname="intns1.zonetransfer.me", dtype=types["A"], ip="81.4.108.41"},
+  {dname="intns2.zonetransfer.me", dtype=types["A"], ip="167.88.42.94"},
+  {dname="office.zonetransfer.me", dtype=types["A"], ip="4.23.39.254"},
+  {dname="ipv6actnow.org.zonetransfer.me", dtype=types["AAAA"], ipv6="2001:67c:2e8:11::c100:1332"},
+  {dname="owa.zonetransfer.me", dtype=types["A"], ip="207.46.197.32"},
+  {dname="robinwood.zonetransfer.me", dtype=types["TXT"], text="Robin Wood"},
+  {dname="rp.zonetransfer.me", dtype=types["RP"]}, -- robin.zonetransfer.me. robinwood.zonetransfer.me.
+  {dname="sip.zonetransfer.me", dtype=types["NAPTR"]}, -- 2 3 "P" "E2U+sip" "!^.*$!sip:customer-service@zonetransfer.me!" .
+  {dname="sqli.zonetransfer.me", dtype=types["TXT"], text="' or 1=1 --"},
+  {dname="sshock.zonetransfer.me", dtype=types["TXT"],text="() { :]}; echo ShellShocked"},
+  {dname="staging.zonetransfer.me", dtype=types["CNAME"], domain="www.sydneyoperahouse.com"},
+  {dname="alltcpportsopen.firewall.test.zonetransfer.me", dtype=types["A"], ip="127.0.0.1"},
+  {dname="testing.zonetransfer.me", dtype=types["CNAME"], domain="www.zonetransfer.me"},
+  {dname="vpn.zonetransfer.me", dtype=types["A"], ip="174.36.59.154"},
+  {dname="www.zonetransfer.me", dtype=types["A"], ip="5.196.105.14"},
+  {dname="xss.zonetransfer.me", dtype=types["TXT"], text="'><script>alert('Boo')</script>"},
+  {dname="zonetransfer.me", dtype=types["SOA"], mname="nsztm1.digi.ninja", rname="robin.digi.ninja"},
+}
+for i, a in ipairs(axfr_decoded.answers) do
+  ta = answers[i]
+  if ta.dtype == types.TXT then
+    test_suite:add_test(unittest.equal(a.dname, ta.dname), i .. ".dname")
+    test_suite:add_test(unittest.equal(a.dtype, ta.dtype), i .. ".dtype")
+    test_suite:add_test(unittest.equal(a.TXT.text[1], ta.text), i .. ".text")
+  elseif ta.dtype == types.SOA then
+    test_suite:add_test(unittest.equal(a.dname, ta.dname), i .. ".dname")
+    test_suite:add_test(unittest.equal(a.dtype, ta.dtype), i .. ".dtype")
+    test_suite:add_test(unittest.equal(a.SOA.mname, ta.mname), i .. ".mname")
+    test_suite:add_test(unittest.equal(a.SOA.rname, ta.rname), i .. ".rname")
+  elseif ta.dtype == types.MX then
+    test_suite:add_test(unittest.equal(a.dname, ta.dname), i .. ".dname")
+    test_suite:add_test(unittest.equal(a.dtype, ta.dtype), i .. ".dtype")
+    test_suite:add_test(unittest.equal(a.MX.pref, ta.pref), i .. ".pref")
+    test_suite:add_test(unittest.equal(a.MX.server, ta.server), i .. ".server")
+  elseif ta.dtype == types.SRV then
+    test_suite:add_test(unittest.equal(a.dname, ta.dname), i .. ".dname")
+    test_suite:add_test(unittest.equal(a.dtype, ta.dtype), i .. ".dtype")
+    test_suite:add_test(unittest.equal(a.SRV.prio, ta.prio), i .. ".prio")
+    test_suite:add_test(unittest.equal(a.SRV.weight, ta.weight), i .. ".weight")
+    test_suite:add_test(unittest.equal(a.SRV.port, ta.port), i .. ".port")
+  else
+    for k, v in pairs(ta) do
+      test_suite:add_test(unittest.equal(a[k], v), ("%d.%s"):format(i, k))
+    end
+  end
+end
 return _ENV;
